__version__ = '1.0.0'

import sys, traceback

from typing import Dict, Optional
import pandas as pd
from datetime import datetime

from beancount.ingest import importer
from beancount.core import data, flags
from beancount.core.amount import Amount
from beancount.core.number import Decimal
from beancount.core.position import Cost


class BDImporter(importer.ImporterProtocol):
    """Beancount Importer for Caisse d'Epargne PDF and CSV statement.

    Attributes:
        account (str): Account name in beancount format (e.g. 'Assets:FR:BD:PEA')
        tickers_lut (dict[str:str]): Look-up tables of the order labels (e.g. {'AM.E.P.SP500':'SP500', ...})
        fee_category (str): Category in which to record fees (e.g. 'Expenses:Finances:Commission')
    """

    def __init__(
        self,
        account: str,
        tickers_lut: Dict[str, str] = dict(),
        fee_category: str = "Expenses:FIXME",
    ):
        self.account = account
        self.tickers_lut = tickers_lut
        self.fee_category = fee_category
        self.expected_columns = {
            'Cours',
            'Date affectation',
            'Date opération',
            'Libellé',
            'Montant net',
            'Opération',
            'Qté',
        }

    def _fetch_ticker(self, asset_name: str) -> Optional[str]:
        matches = []
        for name, ticker in self.tickers_lut.items():
            if name in asset_name:
                matches.append(ticker)
        if len(matches) == 1:
            return matches[0]
        elif len(matches) > 1:
            raise RuntimeError(f"Multiple asset name matched: {matches}")
        elif len(matches) == 0:
            return None

    ## API Methods ##
    #################

    def name(self):
        return 'Bourse Direct: {}'.format(self.__class__.__name__)

    def file_account(self, _):
        return self.account

    def file_date(self, file_):
        if not self.identify(file_):
            return None

        n = file_ if type(file_) == str else file_.name
        date = None
        bd_tables = pd.read_html(n)

        for bf_df in bd_tables:
            if "Date affectation" in set(bf_df.columns):
                for _, row in bf_df.iterrows():
                    try:
                        date_tmp = datetime.strptime(
                            str(row["Date affectation"]), '%d/%m/%Y'
                        ).date()
                    except ValueError:
                        continue
                    if not date or date_tmp > date:
                        date = date_tmp
        return date

    def file_name(self, _):
        return 'BourseDirect.html'

    def identify(self, file_) -> bool:
        n = file_ if type(file_) == str else file_.name
        try:
            bd_tables = pd.read_html(n)
            for df in bd_tables:
                if set(df.columns) == self.expected_columns:
                    break
            else:
                return False
        except:
            return False
        return True

    def extract(self, file_, existing_entries=None):

        entries = (
            list(existing_entries[:]) if existing_entries is not None else []
        )
        n = file_ if type(file_) == str else file_.name

        bd_tables = pd.read_html(n)
        for bf_df in bd_tables:
            if set(bf_df.columns) == self.expected_columns:
                for index, row in bf_df.iterrows():

                    if (
                        isinstance(row['Opération'], str)
                        and "ACHAT" in row['Opération']
                    ):

                        try:
                            op_date = datetime.strptime(
                                row["Date affectation"], '%d/%m/%Y'
                            ).date()
                            op_payee = f"{row['Opération']}: {int(row['Qté'])}x {row['Libellé']}"
                            op_narration = ""

                            op_currency = 'EUR'
                            op_net_amount = Amount(
                                Decimal(
                                    row['Montant net']
                                    .replace("€", "")
                                    .replace(",", ".")
                                    .replace(" ", "")
                                ),
                                op_currency,
                            )
                            op_asset_ticker = self._fetch_ticker(
                                row['Libellé']
                            )
                            op_asset_qtt = Amount(
                                Decimal(
                                    row['Qté'],
                                ),
                                op_asset_ticker
                                if op_asset_ticker is not None
                                else "FIXME",
                            )
                            op_asset_amount = Decimal(
                                row['Cours']
                                .replace("€", "")
                                .replace(",", ".")
                                .replace(" ", ""),
                            )
                            op_asset_price = Amount(
                                op_asset_amount,
                                "EUR",
                            )
                            op_asset_cost = Cost(
                                op_asset_amount, "EUR", None, None
                            )
                        except Exception as e:
                            sys.stderr.write(
                                f"WARNING - Row {index} skipped:\n{traceback.format_exc()}\n"
                            )
                            continue

                        postings = [
                            data.Posting(
                                self.account,
                                op_net_amount,
                                None,
                                None,
                                None,
                                None,
                            ),
                            data.Posting(
                                account=self.account,
                                units=op_asset_qtt,
                                cost=op_asset_cost,
                                price=op_asset_price,
                                flag=None,
                                meta=None,
                            ),
                            data.Posting(
                                self.fee_category,
                                None,
                                None,
                                None,
                                None,
                                None,
                            ),
                        ]

                        entries.append(
                            data.Transaction(
                                meta=data.new_metadata(n, index),
                                date=op_date,
                                flag=flags.FLAG_OKAY,
                                payee=op_payee,
                                narration=op_narration,
                                tags=data.EMPTY_SET,
                                links=data.EMPTY_SET,
                                postings=postings,
                            )
                        )

                    elif (
                        isinstance(row['Opération'], str)
                        and "VIRT" in row['Opération']
                    ):
                        try:
                            op_date = datetime.strptime(
                                row["Date affectation"], '%d/%m/%Y'
                            ).date()
                            op_payee = str(row['Opération'])
                            op_narration = ""
                            op_currency = 'EUR'
                            op_net_amount = Amount(
                                Decimal(
                                    row['Montant net']
                                    .replace("€", "")
                                    .replace(",", ".")
                                    .replace(" ", "")
                                ),
                                op_currency,
                            )
                        except Exception as e:
                            sys.stderr.write(
                                f"WARNING - Row {index} skipped:\n{traceback.format_exc()}\n"
                            )
                            continue

                        postings = [
                            data.Posting(
                                account=self.account,
                                units=op_net_amount,
                                cost=None,
                                price=None,
                                flag=None,
                                meta=None,
                            ),
                            data.Posting(
                                account="Assets:FIXME",
                                units=-op_net_amount,
                                cost=None,
                                price=None,
                                flag=None,
                                meta=None,
                            ),
                        ]

                        entries.append(
                            data.Transaction(
                                meta=data.new_metadata(n, index),
                                date=op_date,
                                flag=flags.FLAG_OKAY,
                                payee=op_payee,
                                narration=op_narration,
                                tags=data.EMPTY_SET,
                                links=data.EMPTY_SET,
                                postings=postings,
                            )
                        )

                    else:
                        sys.stderr.write(
                            f"WARNING - Row {index} skipped: operation '{row['Opération']}' not supported.\n"
                        )
                        continue

        return entries



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['beancount_bd']

package_data = \
{'': ['*']}

install_requires = \
['beancount>=2.2,<3.0', 'pandas>=1.4.3,<2.0.0']

setup_kwargs = {
    'name': 'beancount-bd',
    'version': '1.0.0',
    'description': 'Beancount order history importer for Bourse Direct brooker',
    'long_description': '# Beancount Bourse Direct Importer\n\n[![GitHub Workflow Status](https://img.shields.io/github/workflow/status/ArthurFDLR/beancount-bd/beancount-bd?style=for-the-badge)](https://github.com/ArthurFDLR/beancount-bd/actions)\n[![PyPI](https://img.shields.io/pypi/v/beancount-bd?style=for-the-badge)](https://pypi.org/project/beancount-bd/)\n[![PyPI - Version](https://img.shields.io/pypi/pyversions/beancount-bd.svg?style=for-the-badge)](https://pypi.org/project/beancount-bd/)\n[![GitHub](https://img.shields.io/github/license/ArthurFDLR/beancount-bd?style=for-the-badge)](https://github.com/ArthurFDLR/beancount-bd/blob/master/LICENSE.txt)\n[![Linting](https://img.shields.io/badge/code%20style-black-000000.svg?style=for-the-badge)](https://github.com/psf/black)\n\n`beancount-bd` provides an order history importer for the brooker [Bourse Direct](http://www.boursedirect.fr) to the [Beancount](http://furius.ca/beancount/) format.\n\n## Installation\n\n```console\n    $ pip install beancount-bd\n```\n\n## Usage\n\nAdd ```BDImporter``` to your [Beancount importers config file](https://beancount.github.io/docs/importing_external_data.html#configuration).\n\n```python\nCONFIG = [\n    BDImporter(\n        account=\'Assets:FR:BD:PEA\',\n        fee_category=\'Expenses:Finances:Commission\',\n        tickers_lut={\n            "AM.E.P.SP500": "PE500",\n            "LY.PEANASD": "PUST",\n            "MSC.EM": "PAEEM",\n            "MSC.EUR": "PCEU",\n        }\n    ),\n]\n```\n\n## Contribution\n\nFeel free to contribute!\n\nPlease make sure you have Python 3.6+ and [`Poetry`](https://poetry.eustace.io/) installed.\n\n1. Git clone the repository - `git clone https://github.com/ArthurFDLR/beancount-bd`\n\n2. Install the packages required for development - `poetry install`\n\n3. That\'s basically it. You should now be able to run lint checks and the test suite - `make lint test`.\n',
    'author': 'Arthur Findelair',
    'author_email': 'arthfind@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ArthurFDLR/beancount-bd',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
